#!/usr/bin/env python3

# APSecret ACR - Test Module verifying locally cached secrets to the APLinux ACR get refreshed.

import os
import errno
import subprocess


# a dictionary of 'secret' names and the command to refresh them.
enrolled_secrets = {
    "ACRConnectionData": "/ap/app/aptools/apsecret_refresh_acr.py",

    "Geneva-HotPath-Cert": "/ap/app/aptools/apsecret_refresh_geneva_hot.py -e prod -d /ap/data/aplinux_geneva/mdm",
    "Geneva-HotPath-Key": "/ap/app/aptools/apsecret_refresh_geneva_hot.py -e prod -d /ap/data/aplinux_geneva/mdm",
    "GenevaINT-HotPath-Cert": "/ap/app/aptools/apsecret_refresh_geneva_hot.py -e int -d /ap/data/aplinux_geneva/mdm",
    "GenevaINT-HotPath-Key": "/ap/app/aptools/apsecret_refresh_geneva_hot.py -e int -d /ap/data/aplinux_geneva/mdm",

    # Commented test items do not yet have a key associated with them since they are not configured in Geneva yet
    # "Geneva-APLinuxGeneva-autokey": "/ap/app/aptools/apsecret_refresh_geneva_warm.py -e prod -n aplinuxgeneva -d /tmp",
    "Geneva-AzureSecurityPack-autokey": "/ap/app/aptools/apsecret_refresh_geneva_warm.py -e prod -n azsecpack -d /ap/data/aplinux_geneva/mdsd_aplinuxazsecpack",
    "Geneva-Syslog-autokey": "/ap/app/aptools/apsecret_refresh_geneva_warm.py -e prod -n syslog -d /ap/data/aplinux_geneva/mdsd_syslog",
    "Geneva-PrivateKeyPFX": "/ap/app/aptools/apsecret_refresh_geneva_warm.py -e prod -n syslog -d /ap/data/aplinux_geneva/mdsd_syslog",
    "Geneva-PrivateKeyPFX-Password": "/ap/app/aptools/apsecret_refresh_geneva_warm.py -e prod -n syslog -d /ap/data/aplinux_geneva/mdsd_syslog",
    # "GenevaINT-APLinuxGeneva-autokey": "/ap/app/aptools/apsecret_refresh_geneva_warm.py -e int -n aplinuxgeneva -d /tmp",
    "GenevaINT-AzureSecurityPack-autokey": "/ap/app/aptools/apsecret_refresh_geneva_warm.py -e int -n azsecpack -d /ap/data/aplinux_geneva/mdsd_aplinuxazsecpack",
    "GenevaINT-Syslog-autokey": "/ap/app/aptools/apsecret_refresh_geneva_warm.py -e int -n syslog -d /ap/data/aplinux_geneva/mdsd_syslog",
    "GenevaINT-PrivateKeyPFX": "/ap/app/aptools/apsecret_refresh_geneva_warm.py -e int -n syslog -d /ap/data/aplinux_geneva/mdsd_syslog",
    "GenevaINT-PrivateKeyPFX-Password": "/ap/app/aptools/apsecret_refresh_geneva_warm.py -e int -n syslog -d /ap/data/aplinux_geneva/mdsd_syslog"
}


def delete_file(path: str) -> bool:
    try:
        os.remove(path)
    except OSError as e:
        if e.errno != errno.ENOENT:  # errno.ENOENT = no such file or directory
            return False  # a different error occured
    return True


def construct_secret_path(name: str) -> str:
    return "/var/cache/apsecret/secrets/" + name


# delete cache, refresh, check cache
def test_apsecret_cache():
    # test all secrets
    for name, refresh_script in enrolled_secrets.items():
        path = construct_secret_path(name)

        # delete the cached file
        assert delete_file(path), "internal error in BVT script: unable to delete file '{0}' for secret '{1}".format(path, name)

        # call the refresh script
        try:
            result = subprocess.run(refresh_script,
                                    timeout=15,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE,
                                    universal_newlines=True,
                                    shell=True,
                                    check=False)
        except subprocess.TimeoutExpired:
            raise Exception("timeout (15s) running '{0}' for refresh of secret '{1}'".format(refresh_script, name))
        except subprocess.CalledProcessError:
            raise Exception("error running '{0}' for refresh of secret '{1}'".format(refresh_script, name))

        assert result.returncode == 0, "'{0}', refresh secret '{1}': non-zero return code"
        assert result.stderr == "", "'{0}', refresh secret '{1}': unexpected output printed to STDERR: '{2}'".format(refresh_script, name, result.stderr.replace("\n", ""))

        # check cache has contents
        assert os.path.isfile(path), "secret '{0}' not downloaded to cache".format(name)
        assert os.path.getsize(path) > 0, "secret '{0}' is empty".format(name)

        # empty the cached file, this is to ensure an existing cache file is overwritten
        try:
            open(path, "w").close()
        except Exception:
            raise Exception("internal error in BVT script: unable to create an empty file for "
                            "secret '{0}'".format(name))

        # ensure the cached file is now zero-length
        assert os.path.isfile(path), "internal error in BVT script: unable to write an empty cache file for secret '{0}'".format(name)
        assert os.path.getsize(path) == 0, "internal error in BVT script: unable to delete the contents of secret '{0}'".format(name)

        # call the refresh script
        try:
            result = subprocess.run(refresh_script,
                                    timeout=15,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE,
                                    universal_newlines=True,
                                    shell=True,
                                    check=False)
        except subprocess.TimeoutExpired:
            raise Exception("timeout (15s) running '{0}' for refresh of secret '{1}'".format(refresh_script, name))
        except subprocess.CalledProcessError:
            raise Exception("error running '{0}' for refresh of secret '{1}'".format(refresh_script, name))

        assert result.returncode == 0, "'{0}', refresh secret '{1}': non-zero return code".format(refresh_script, name)
        assert result.stderr == "", "'{0}', refresh secret '{1}': unexpected output printed to STDERR: '{2}'".format(refresh_script, name, result.stderr.replace("\n", ""))

        # check cache has contents
        assert os.path.isfile(path), "secret '{0}' not downloaded to cache".format(name)
        assert os.path.getsize(path) > 0, "secret '{0}' is empty".format(name)
