#!/usr/bin/env python3

# Deployment - Test Module verifying deb package installation

import glob
import os
import shutil
import string
import random
import time

from common.appath import APPath
from common.shellhelper import ShellHelper
from distutils.dir_util import copy_tree


# ---------- Module Test Fixtures -----------
# Run once for each module
def setup_module(module):
    setup_test_data()


def teardown_module(module):
    cleanup_test_data()


# This is run before each test case function is executed. Have here for enabling later.
def setup_function(function):
    pass


def teardown_function(function):
    pass


# ---------- End Fixtures -----------

# Deb Test helpers
def setup_test_data():
    print("Copying test packages to packed images dir")
    try:
        copy_tree("./deployment/test_data/", "/ap/app/packedimages/")
    except shutil.Error as e:
        print('Test setup error :' + str(e))


def setup_deb_test_package():
    test_svclist = get_service_list_name('TEST_DEB')
    print("Test Service List Name:", test_svclist)
    print("Copying test packages to packed images dir")
    try:
        copy_tree("./deployment/test_data/TEST_DEB.csv", "/ap/app/packedimages/" + test_svclist)
    except shutil.Error as e:
        print('Test setup error :' + str(e))

    return test_svclist


def cleanup_test_data():
    try:
        for path in glob.glob("/ap/app/packedimages/TEST_*"):
            print('Removing', path)
            shutil.rmtree(path)
    except OSError as e:
        print('Test setup error :' + str(e))


def remove_pkgs(pkg_list):
    for pkg in pkg_list:
        ret, result = ShellHelper.run_cmd_output("apt-get -qq -y -f --purge remove " + pkg, 30, 4)
        assert ret, "Could not remove " + pkg


def verify_pkg_ver(pkgname, pkgver):
    cmd = "dpkg-query --showformat='${Version}' --show " + pkgname

    ret, result = ShellHelper.run_cmd_output(cmd, 30, 2)
    assert ret, "command execution failed" + cmd
    assert result == pkgver, "Package version did not match target version: " + pkgname + "=" + pkgver


def verify_commands(cmd_list):
    for cmd in cmd_list:
        ret, result = ShellHelper.run_cmd_output(cmd, 30, 2)
        assert ret, "Verifying command failed: " + cmd


def wait_for_completion(servicelist, test_svclist, expectError=False):
    print("Monitor service manager folder for error package.")
    for i in range(0, 30):
        # Sleep before polling the service manager folder for error.
        # Adjust if your pkg processing takes longer and log lines are too verbose.
        time.sleep(3)
        # Check if an error file is generated for this package.
        for f in glob.glob(APPath.ServiceManagerDataRoot + "*.err"):
            if expectError is False:
                assert test_svclist not in f, "Service list processing failed unexpectedly: " + test_svclist
            else:
                if test_svclist in f:
                    # All good, we got the expected error
                    print("Expected error file found:", f)
                    return f

        # Check if active_servicelist.ini is updated. processing completed.
        if servicelist.verify_active_list(test_svclist):
            print("Service list processing completed" + test_svclist)
            assert expectError is False, "Service list processing error expected, but passed: " + test_svclist
            return

    # If we reach here, we neither passed or failed within 90 seconds. Package processing
    # should complete within this time. If not consider this a failure.
    assert False, "Service list processing timed out. No result came back: " + test_svclist


def random_id(size=6, chars=string.ascii_uppercase + string.digits):
    return ''.join(random.choice(chars) for _ in range(size))


def get_service_list_name(prefix):
    return prefix + '_' + random_id() + '.csv'


def create_service_list_folder(service_list_name):
    os.mkdir(APPath.APPackageFolder + service_list_name)


def get_var_file_name(service_list_name):
    return APPath.APPackageFolder + service_list_name + '/install_deb_pkgs/vars/deb_list_var.yml'


def find_file(pattern):
    for f in glob.glob(pattern):
        assert os.path.getsize(f), "File exists, but is empty : " + f
        return True

    # Didn't find any file of pattern specified
    return False


def verify_debug_output(service_list_name, success=True):
    log_root = os.path.join("/ap/data/logs/rollout", service_list_name)

    # Check for ansible output link
    assert find_file(log_root + "/*.ansible.log"), "Ansible output file missing in : " + log_root

    # Check for error file, if this is a failure
    if not success:
        assert find_file(log_root + "/*.failure.log"), "Ansible error file missing in : " + log_root
    else:
        assert find_file(
            log_root + "/*.failure.log") is False, "Ansible error file exists on success : " + log_root


# Write debian variable file in yaml format.
def write_deb_var_file(filepath, pkg_and_ver):
    content = """---
# Variables for installing debian packages

# TEST: List of packages to be installed and their version. Version can be 'latest' or a
# specific version
debian_pkgs:"""
    with open(filepath, 'w+') as f:
        f.write(content)
        for k, v in pkg_and_ver.items():
            f.write("\n    - name: " + k)
            f.write("\n      version: " + v + "\n")


# Test Cases. Should start with test_ to be picked by pytest.
def test_deploy_latest(apdeployment, servicemanager, servicelist):
    # Test installing latest, pkg with specific versions and already installed packages.
    pkg_and_ver = {'finger': 'latest',
                   'cowsay': 'latest',
                   'sl': 'latest',
                   'rig': 'latest',
                   'python-ogg': 'latest'}

    # Pause system deployment pipeline
    apdeployment.pause()

    # setup_test_data()
    test_svclist = setup_deb_test_package()
    servicelist.update_list(test_svclist)
    write_deb_var_file(get_var_file_name(test_svclist), pkg_and_ver)

    # Remove the packages
    pkg_list = ['finger',
                'cowsay',
                'sl',
                'rig',
                'python-ogg']

    remove_pkgs(pkg_list)

    print("Starting service manager to process the package")
    servicemanager.restart()

    wait_for_completion(servicelist, test_svclist)
    # Verify that no error message is written in JSON.
    # Enable this once this feature from master is rolled out.
    # assert not servicelist.get_active_error_message(), "ErrorMessage not found in active_servicelist.json"

    servicemanager.restore_state()

    print("Confirm the debian packages are installed")
    commands = ['dpkg-query -W finger',
                'dpkg-query -W cowsay',
                'dpkg-query -W rig',
                'dpkg-query -W python-ogg',
                'dpkg-query -W sl']
    verify_commands(commands)

    verify_debug_output(test_svclist)

    remove_pkgs(pkg_list)


def test_deploy_specific_version(apdeployment, servicemanager, servicelist):
    # Test installing specific versions of packages
    pkg_and_ver = {  # Specific version of the package
        'fortune-mod': '1:1.99.1-7',
        'figlet': '2.2.5-2',
        'toilet': '0.3-1'}

    # Pause system deployment pipeline
    apdeployment.pause()

    # setup_test_data()
    test_svclist = setup_deb_test_package()
    servicelist.update_list(test_svclist)
    write_deb_var_file(get_var_file_name(test_svclist), pkg_and_ver)

    # Remove the packages
    pkg_list = ['fortune-mod',
                'figlet',
                'toilet']

    remove_pkgs(pkg_list)

    print("Starting service manager to process the package")
    servicemanager.restart()

    wait_for_completion(servicelist, test_svclist)
    # Verify that no error message is written in JSON.
    # Enable this once this feature from master is rolled out.
    # assert not servicelist.get_active_error_message(), "ErrorMessage not found in active_servicelist.json"

    servicemanager.restore_state()

    print("Confirm the debian packages are installed")
    # Package installed or timed out.
    commands = ['/usr/games/fortune',
                'which figlet']
    verify_commands(commands)

    verify_debug_output(test_svclist)

    # Verify specific target versions of the pkgs are installed
    verify_pkg_ver("fortune-mod", "1:1.99.1-7")
    verify_pkg_ver("toilet", "0.3-1")
    verify_pkg_ver("figlet", "2.2.5-2")

    remove_pkgs(pkg_list)


def test_deploy_already_installed(apdeployment, servicemanager, servicelist):
    # Test installing packages that were already installed.
    pkg_and_ver = {'adduser': 'latest',
                   'at': 'latest',
                   'bash': 'latest',
                   'linux-generic': 'latest',
                   'linux-headers-generic': 'latest'}

    # Pause system deployment pipeline
    apdeployment.pause()

    # setup_test_data()
    test_svclist = setup_deb_test_package()
    servicelist.update_list(test_svclist)
    write_deb_var_file(get_var_file_name(test_svclist), pkg_and_ver)

    print("Starting service manager to process the package")
    servicemanager.restart()

    wait_for_completion(servicelist, test_svclist)
    # Verify that no error message is written in JSON.
    # Enable this once this feature from master is rolled out.
    # assert not servicelist.get_active_error_message(), "ErrorMessage not found in active_servicelist.json"

    servicemanager.restore_state()

    print("Confirm the debian packages are there")
    commands = ['which adduser',
                'which at']
    verify_commands(commands)

    verify_debug_output(test_svclist)


def test_failure(apdeployment, servicemanager, servicelist):
    # Test deployment failure is handled properly
    pkg_and_ver = {'adduser': 'wrong-ver.1'}

    # Pause system deployment pipeline
    apdeployment.pause()

    # setup_test_data()
    test_svclist = setup_deb_test_package()
    servicelist.update_list(test_svclist)
    write_deb_var_file(get_var_file_name(test_svclist), pkg_and_ver)

    print("Starting service manager to process the package")
    servicemanager.restart()

    error_file = wait_for_completion(servicelist, test_svclist, True)

    # Verify that an error message is written in JSON on failure.
    # Enable this once this feature from master is rolled out.
    # assert servicelist.get_active_error_message(), "ErrorMessage not found in active_servicelist.json"

    # Remove the error file
    os.remove(error_file)

    servicemanager.restore_state()

    print("Confirm the older debian packages are there")
    commands = ['which adduser']
    verify_commands(commands)

    verify_debug_output(test_svclist, False)


def test_failure_and_rollback(apdeployment, servicemanager, servicelist):
    # Test a deb deployment failure and rollback
    # Make sure we remove the packages
    remove_pkgs(['finger'])

    # Pause system deployment pipeline
    apdeployment.pause()

    # Setup AP package that will pass deployment and initiate rollout. Should pass.
    pkg_and_ver = {'finger': 'latest'}
    test_svclist_v0 = setup_deb_test_package()
    write_deb_var_file(get_var_file_name(test_svclist_v0), pkg_and_ver)

    # active_list = <current-active-list>, set new_list to test_svclist_v0
    servicelist.update_list(test_svclist_v0)
    servicemanager.restart()
    error_file = wait_for_completion(servicelist, test_svclist_v0, False)
    # Ensure we got to the correct active_list
    assert servicelist.get_active_list() == test_svclist_v0, "Active list is still not updated to " + test_svclist_v0
    # State: active_list = test_svclist_v0, new_list = test_svclist_v0
    # ------------------ v0 deployment succeeded ---------------

    # Lets remove the package for rollback verification later
    remove_pkgs(['finger'])

    # Setup AP package that will fail deployment and initiate rollout. Should fail.
    pkg_and_ver = {'finger': 'wrong-ver.1'}
    test_svclist_v1 = setup_deb_test_package()
    write_deb_var_file(get_var_file_name(test_svclist_v1), pkg_and_ver)

    # active_list = test_svclist_v0
    # set new_list to test_svclist_v1
    servicelist.update_list(test_svclist_v1)
    servicemanager.restart()
    error_file = wait_for_completion(servicelist, test_svclist_v1, True)

    # Ensure we got to the correct active_list
    assert servicelist.get_active_list() == test_svclist_v1, "Active list is still not updated to " + test_svclist_v1
    # There should be an error to be sent up
    assert test_svclist_v1 in servicelist.get_active_error_message(), "ErrorMessage not found in active_servicelist.json"

    # Remove the error file
    os.remove(error_file)

    # State: active_list = test_svclist_v1, ErrorMessage="RolloutFailure-xxxxx",
    # new_list = test_svclist_v1
    # ------------------ v1 deployment failed ---------------

    # Setup rollback state.
    # new_list = test_svclist_v0 and active_list = test_svclist_v1
    servicelist.update_list(test_svclist_v0)
    servicemanager.restart()
    wait_for_completion(servicelist, test_svclist_v0, False)
    # Ensure we got to the correct active_list
    assert servicelist.get_active_list() == test_svclist_v0, "On Rollback, active list was not updated to " + test_svclist_v0
    verify_commands(['dpkg-query -W finger'])
