#!/usr/bin/env python3

# BVT for ensuring Docker images used by shared services can be successfully pulled

import subprocess
import time


# a dictionary of image name and the service name to start and stop them
enrolled_containers = {
    "aplinux-microsoft.azurecr.io.apdocker.ap.gbl/apnginx": "apnginx"
}

SERVICE_STOP_TIMEOUT = 60
DELETE_ALL_IMAGES_TIMEOUT = 30
IMAGE_PULL_TIMEOUT = 300
IMAGE_POLLING_INTERVAL = 10


def count_image(name: str) -> int:
    # one line per image tag
    try:
        result = subprocess.run("docker images -q {0}".format(name),
                                timeout=15,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE,
                                universal_newlines=True,
                                shell=True,
                                check=False)
    except subprocess.TimeoutExpired:
        raise Exception("timeout (15s) running 'docker images -q {0}' for counting images".format(name))
    except subprocess.CalledProcessError:
        raise Exception("error running 'docker images -q {0}' for counting images".format(name))

    assert result.returncode == 0, "count image tags for {0}: non-zero return code".format(name)
    assert result.stderr == "", "count image tags for {0}: unexpected output printed to STDERR: '{1}'".format(name, result.stderr.replace("\n", ""))

    return len(result.stdout.split('\n')) - 1


# stop services, delete images, start services
def test_docker_images():
    # stop all services
    for image, service in enrolled_containers.items():
        try:
            subprocess.run("service {0} stop".format(service),
                           timeout=SERVICE_STOP_TIMEOUT,
                           stdout=subprocess.PIPE,
                           stderr=subprocess.PIPE,
                           universal_newlines=True,
                           shell=True,
                           check=False)
        except subprocess.TimeoutExpired:
            raise Exception("timeout ({0}s) running 'service {1} stop' for stopping service".format(SERVICE_STOP_TIMEOUT, service))
        except subprocess.CalledProcessError:
            raise Exception("error running 'service {0} stop' for stopping service".format(service))

    # delete all the Docker images
    for image, service in enrolled_containers.items():
        try:
            subprocess.run("docker rmi $(docker images -q {0})".format(image),
                           timeout=DELETE_ALL_IMAGES_TIMEOUT,
                           stdout=subprocess.PIPE,
                           stderr=subprocess.PIPE,
                           universal_newlines=True,
                           shell=True,
                           check=False)
        except subprocess.TimeoutExpired:
            raise Exception("timeout ({0}s) running 'docker rmi $(docker images -q {1})' for deleting images".format(DELETE_ALL_IMAGES_TIMEOUT, image))
        except subprocess.CalledProcessError:
            raise Exception("error running 'docker rmi $(docker images -q {0})' for deleting images".format(image))

        assert count_image(image) == 0, "failed to delete image '{0}' for service '{1}'".format(image, service)

    # start the service and ensure the docker image is pulled before timeout is reached
    for image, service in enrolled_containers.items():
        # start the service
        try:
            subprocess.run("service {0} start".format(service),
                           timeout=IMAGE_PULL_TIMEOUT,
                           stdout=subprocess.PIPE,
                           stderr=subprocess.PIPE,
                           universal_newlines=True,
                           shell=True,
                           check=False)
        except subprocess.TimeoutExpired:
            raise Exception("timeout ({0}s) running 'service {1} start' for starting service '{1}'".format(IMAGE_PULL_TIMEOUT, service))
        except subprocess.CalledProcessError:
            raise Exception("error running 'service {0} start' for starting service '{0}'".format(service))

        # poll docker until image is available or timeout is exceeded
        poll_time = 0
        while poll_time < IMAGE_PULL_TIMEOUT:
            if count_image(image) > 0:
                print("image '{0}' successfully pulled for service '{1}'".format(image, service))
                break
            time.sleep(IMAGE_POLLING_INTERVAL)
            poll_time = poll_time + IMAGE_POLLING_INTERVAL

        assert poll_time < IMAGE_PULL_TIMEOUT, "image '{0}' is unavailable for service '{1}' after timeout of {2} seconds".format(image, service, IMAGE_PULL_TIMEOUT)
        assert count_image(image) > 0, "image '{0}' is unavailable for service '{1}'".format(image, service)
