#!/usr/bin/env python3

# APSecret ACR - Test Module verifying locally cached secrets to the APLinux ACR get refreshed.

import os
import socket
import subprocess


# intermediate certificate chain
root_certs = "/usr/share/ca-certificates/ap/apca_roots.pem"
intermediate_certs = "/usr/share/ca-certificates/ap/apca_intermediates.pem"

# machine and machine function certificates
machine_private_cert = "/etc/ssl/certs/{0}.key"
machine_public_cert = "/etc/ssl/certs/{0}.crt"
mf_private_cert = "/etc/ssl/certs/{0}.mf.key"
mf_public_cert = "/etc/ssl/certs/{0}.mf.crt"

# disallowed cert store
disallowed_store = "/usr/share/ca-certificates/ap/disallowed.csv"


def get_machine_name() -> str:
    hostname = ""

    try:
        hostname = os.environ["AP_MACHINE_NAME"]
    except:
        try:
            hostname = socket.gethostname()
        except:
            print("Error: No machine name specified and unable to get "
                  "machine hostname")
            hostname = ""

    # make sure hostname is not empty
    assert hostname

    return hostname


# check public certificates can be parsed
def try_parse_public_cert(path: str) -> bool:
    result = None
    command = "openssl x509 -inform PEM -text -noout -in '{0}'".format(path)
    success = True

    try:
        result = subprocess.run(command,
                                timeout=15,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE,
                                universal_newlines=True,
                                shell=True,
                                check=False)
    except subprocess.TimeoutExpired:
        print("timeout (15s) running '{0}' for parsing public "
              "cert file '{1}'".format(command, path))
        return False
    except subprocess.CalledProcessError:
        print("error running '{0}' for parsing public "
              "cert file '{1}'".format(command, path))
        return False

    if result.returncode != 0:
        print("'{0}', parse public cert file '{1}': non-zero "
              "return code ({2})".format(command,
                                         path,
                                         result.returncode))
        success = False

    if result.stderr:
        safe_stderr = result.stderr.replace("\n", " ")
        print("'{0}', parse public cert file '{1}': unexpected output "
              "printed to STDERR: '{2}'".format(command,
                                                path,
                                                safe_stderr))
        success = False

    return success


# check private certificates can be parsed
def try_parse_private_cert(path: str) -> bool:
    result = None
    command = "openssl rsa -check -in '{0}'".format(path)
    success = True

    try:
        result = subprocess.run(command,
                                timeout=15,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE,
                                universal_newlines=True,
                                shell=True,
                                check=False)
    except subprocess.TimeoutExpired:
        print("timeout (15s) running '{0}' for parsing private "
              "cert file '{1}'".format(command, path))
        return False
    except subprocess.CalledProcessError:
        print("error running '{0}' for parsing private "
              "cert file '{1}'".format(command, path))
        return False

    if result.returncode != 0:
        print("'{0}', parse private cert file '{1}': non-zero "
              "return code ({2})".format(command,
                                         path,
                                         result.returncode))
        success = False

    if result.stderr and not (result.stderr.strip() == "writing RSA key"):
        safe_stderr = result.stderr.replace("\n", " ")
        print("'{0}', parse private cert file '{1}': unexpected output "
              "printed to STDERR: '{2}'".format(command,
                                                path,
                                                safe_stderr))
        success = False

    return success


# Returns True if the public and private certificates are a pair
def is_cert_pair(public_cert_path: str, private_cert_path: str) -> bool:
    result = None
    command_pub = "openssl x509 -inform PEM -noout -pubkey " \
                  "-in '{0}'".format(public_cert_path)
    command_pri = "openssl rsa -pubout -in '{0}'".format(private_cert_path)
    command_pub_pubkey = ""
    command_pri_pubkey = ""
    success = True

    # Get pubkey for the public certificate
    try:
        result = subprocess.run(command_pub,
                                timeout=15,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE,
                                universal_newlines=True,
                                shell=True,
                                check=False)
        command_pub_pubkey = result.stdout
    except subprocess.TimeoutExpired:
        print("timeout (15s) running '{0}' for getting pubkey for public "
              "cert file '{1}'".format(command_pub, public_cert_path))
        return False
    except subprocess.CalledProcessError:
        print("error running '{0}' for getting pubkey for public "
              "cert file '{1}'".format(command_pub, public_cert_path))
        return False

    if result.returncode != 0:
        print("'{0}', get pubkey for public cert file '{1}': non-zero "
              "return code ({2})".format(command_pub,
                                         public_cert_path,
                                         result.returncode))
        success = False

    # Get pubkey for the private certificate
    try:
        result = subprocess.run(command_pri,
                                timeout=15,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE,
                                universal_newlines=True,
                                shell=True,
                                check=False)
        command_pri_pubkey = result.stdout
    except subprocess.TimeoutExpired:
        print("timeout (15s) running '{0}' for getting pubkey for private "
              "cert file '{1}'".format(command_pri, private_cert_path))
        return False
    except subprocess.CalledProcessError:
        print("error running '{0}' for getting pubkey for private "
              "cert file '{1}'".format(command_pri, private_cert_path))
        return False

    if result.returncode != 0:
        print("'{0}', get pubkey for private cert file '{1}': non-zero "
              "return code ({2})".format(command_pri,
                                         private_cert_path,
                                         result.returncode))
        success = False

    # Get out of here if one of the openssl commands failed or didn't output
    if not success or not command_pub_pubkey or not command_pri_pubkey:
        print("Openssl commands failed to run successfully. Did you get the "
              "certificate types correct? Must be in PEM format, public cert "
              "must be a public cert, private cert must be a private cert")
        return False

    # Return True if pubkey is the same
    return command_pub_pubkey == command_pri_pubkey


def is_cert_trusted(path: str) -> bool:
    result = None
    command = "openssl verify -CAfile '{0}' -untrusted '{1}' " \
              "'{0}'".format(root_certs,
                             intermediate_certs,
                             path)
    success = True

    try:
        result = subprocess.run(command,
                                timeout=15,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE,
                                universal_newlines=True,
                                shell=True,
                                check=False)
    except subprocess.TimeoutExpired:
        print("timeout (15s) running '{0}' for verifying trust of "
              "cert file '{1}'".format(command, path))
        return False
    except subprocess.CalledProcessError:
        print("error running '{0}' for verifying trust of "
              "cert file '{1}'".format(command, path))
        return False

    if result.returncode != 0:
        print("'{0}', verify trust of cert file '{1}': non-zero "
              "return code".format(command, path))
        success = False

    if result.stderr:
        safe_stderr = result.stderr.replace("\n", " ")
        print("'{0}', verify trust of cert file '{1}': unexpected output "
              "printed to STDERR: '{2}'".format(command,
                                                path,
                                                safe_stderr))
        success = False

    return success


# check the file and make sure its a machine certificate
def is_machine_certificate(path: str) -> bool:
    # X509v3 Subject Alternative Name:
    #   DNS:CO4PHY070012701.CO4
    return True


# check the file and make sure its a machine function certificate
def is_mf_certificate(path: str) -> bool:
    # X509v3 Subject Alternative Name:
    #   DNS:CO4PHY070012701.CO4, DNS:LINUXTEST.Azlinux-Test-CO4.CO4
    return True


# check file validity and certificate trust
def test_certificates():
    hostname = get_machine_name()

    # make sure our cert chain is valid
    assert try_parse_public_cert(root_certs)
    assert try_parse_public_cert(intermediate_certs)

    # test machine and machine function certificates
    assert try_parse_public_cert(machine_public_cert.format(hostname))
    assert try_parse_private_cert(machine_private_cert.format(hostname))
    assert try_parse_public_cert(mf_public_cert.format(hostname))
    assert try_parse_private_cert(mf_private_cert.format(hostname))

    # test to see if public key and private key are a pair (belong together)
    assert is_cert_pair(machine_public_cert.format(hostname),
                        machine_private_cert.format(hostname))
    assert is_cert_pair(mf_public_cert.format(hostname),
                        mf_private_cert.format(hostname))

    # test to see if our machine and machine function certificates are trusted
    assert is_cert_trusted(machine_public_cert.format(hostname))
    assert is_cert_trusted(mf_public_cert.format(hostname))

    # verify the machine certificate is a machine certificate and the
    # machine function certificate is a machine function certificate
    assert is_machine_certificate(machine_public_cert.format(hostname))
    assert is_mf_certificate(mf_public_cert.format(hostname))


# check the disallowed certificate store
def test_disallowed_store():
    # make sure the file exists and is not empty
    assert os.path.exists(disallowed_store)
    assert os.stat(disallowed_store).st_size > 0
